
//***************************************************
// Generalised Bitmap Module C++ adapter for GBM plugins
//
// Requires: GBM.DLL version 1.35 or higher (with multipage support)
//
// Copyright (C) 2006 Heiko Nitzsche
//
//   This software is provided 'as-is', without any express or implied
//   warranty.  In no event will the author be held liable for any damages
//   arising from the use of this software.
//
//   Permission is granted to anyone to use this software for any purpose,
//   including commercial applications, and to alter it and redistribute it
//   freely, subject to the following restrictions:
//
//   1. The origin of this software must not be misrepresented; you must not
//      claim that you wrote the original software. If you use this software
//      in a product, an acknowledgment in the product documentation would be
//      appreciated but is not required.
//   2. Altered source versions must be plainly marked as such, and must not be
//      misrepresented as being the original software.
//   3. This notice may not be removed or altered from any source distribution.
//
//***************************************************

#ifdef __IBMCPP__
  #pragma strings( readonly )
#endif

#include "os2defs.h"

#include <string.h>
#include <stdlib.h>

#include "GbmRenderer.hpp"

#include "gbm.h"
#include "gbmscale.h"
#include "gbmmir.h"
#include "gbmrect.h"

// ---------------------------------------------------------

GbmRenderer::GbmRenderer()
    : renderPage(NULL),
      unscaledWidth(0),
      unscaledHeight(0)
{
}

// ---------------------------------------------------------

GbmRenderer::~GbmRenderer()
{
    renderPage = NULL;
}

// ---------------------------------------------------------

void GbmRenderer::setRenderPage(GbmDocumentPage * page)
{
    if (page != renderPage)
    {
        // clear old render cache
        cacheDescription.clear();
        renderPage = page;
    }
}

GbmDocumentPage * GbmRenderer::getRenderPage() const
{
    return renderPage;
}

// ---------------------------------------------------------

/**
 * @param src_x         x coordinate of upper left corner
 * @param src_y         y coordinate of upper left corner
 * @param src_width     width of rectangle to render
 * @param src_height    height of rectangle to render
 * @param scale_x       scale specified as pixels per point (width)
 * @param scale_y       scale specified as pixels per point (height)
 * @param renderBuffer  render buffer (4 byte aligned row size according
 *                      to src_width and page colour depth of page set)
 *
 * 1. Scale the document to match the specified pixels per point
 * 2. Render the rectangle given by the upper left corner at
 *    (src_x, src_y) and src_width and src_height.
 **/

void GbmRenderer::renderToBuffer(long     src_x    , long   src_y,
                                 long     src_width, long   src_height,
                                 double   scale_x  , double scale_y,
                                 byte   * renderBuffer) throw ( GbmException )
{
    if (renderPage == NULL)
    {
        throw GbmException(GBM_ERR_BAD_ARG, "No render page set");
    }

    if ((scale_x < 0.0) || (scale_y < 0.0) ||
        (src_y < 0)     || (src_x < 0)     || (src_width <= 0) || (src_height <= 0))
    {
        throw GbmException(GBM_ERR_BAD_ARG, "Bad render options");
    }

    // Check if any important render parameters (scaling, page size) have changed.
    if ((scale_x                 != cacheDescription.scale_x) ||
        (scale_y                 != cacheDescription.scale_y) ||
        (renderPage->getWidth()  != unscaledWidth)            ||
        (renderPage->getHeight() != unscaledHeight))
    {
        // force full recalculation of target bitmap
        cacheDescription.clear();
        unscaledWidth  = renderPage->getWidth();
        unscaledHeight = renderPage->getHeight();
    }

          int    copySrcWidth  = cacheDescription.width;
          int    copySrcHeight = cacheDescription.height;
          int    copySrcStride = cacheDescription.rowStride;
    const byte * copySrcData   = cacheDescription.getDataPtr();
          byte * scaledData    = NULL;

    if (copySrcData == NULL)
    {
        // the page is already rotate if this was necessary
        copySrcWidth  = renderPage->getWidth();
        copySrcHeight = renderPage->getHeight();
        copySrcStride = renderPage->getBitmapDataRowSize();
        copySrcData   = renderPage->getBitmapDataPtr();

        // Needs additional scaling ?

        // check what is more memory conservative (scale first or rotate first)
        if ((scale_x != 1.0) || (scale_y != 1.0))
        {
            const int scaledWidth  = (int) (scale_x * ((double) (copySrcWidth )));
            const int scaledHeight = (int) (scale_y * ((double) (copySrcHeight)));
            const int scaledStride = ((scaledWidth * renderPage->getColorDepth() + 31)/32) * 4;

            scaledData = new byte[scaledStride * scaledHeight];
            if (scaledData != NULL)
            {
                // scale to requested size
                if (gbm_simple_scale(copySrcData, copySrcWidth , copySrcHeight,
                                     scaledData , scaledWidth, scaledHeight,
                                     renderPage->getColorDepth()) == GBM_ERR_OK)
                {
                    // set the copy source to the scaled data
                    copySrcWidth  = scaledWidth;
                    copySrcHeight = scaledHeight;
                    copySrcStride = scaledStride;
                    copySrcData   = scaledData;
                }
            }
            else
            {
                /* In case we had a problem with scaling (e.g. out of memory), */
                /* render nothing (or later show an error icon?).              */
                throw GbmException(GBM_ERR_MEM);
            }
        }

        cacheDescription.width     = copySrcWidth;
        cacheDescription.height    = copySrcHeight;
        cacheDescription.rowStride = copySrcStride;
        cacheDescription.scale_x   = scale_x;
        cacheDescription.scale_y   = scale_y;

        if (copySrcData == scaledData)
        {
            // this data can be freed as we own it
            cacheDescription.setData((byte *) scaledData);
        }
        else
        {
            // just store the reference which will not be freed
            cacheDescription.setConstData(copySrcData);

            // Don't free the scaledData buffer if it points
            // to the same address as the cache.
            delete [] scaledData;
            scaledData = NULL;
        }
    }

    // copy visible rectangle to the drawing buffer
    GBM renderGbm;
    renderGbm.w   = cacheDescription.width;
    renderGbm.h   = cacheDescription.height;
    renderGbm.bpp = renderPage->getColorDepth();

    // protect against buffer overrun due to rounding errors
    if (src_x + src_width > renderGbm.w)
    {
        src_width = renderGbm.w - src_x;
    }
    if (src_y + src_height > renderGbm.h)
    {
        src_height = renderGbm.h - src_y;
    }

    gbm_subrectangle(&renderGbm,
                      src_x, renderGbm.h-(src_y+src_height), src_width, src_height,
                      cacheDescription.getDataPtr(), renderBuffer);

}


